/*
 * String manipulation library for Boehm GC
 *
 * (C) Copyright 1998-1999 by Akinori Ito
 *
 * This software may be redistributed freely for this purpose, in full 
 * or in part, provided that this entire copyright notice is included 
 * on any copies of this software and applications and derivations thereof.
 *
 * This software is provided on an "as is" basis, without warranty of any
 * kind, either expressed or implied, as to any matter including, but not
 * limited to warranty of fitness of purpose, or merchantability, or
 * results obtained from use of this software.
 */
#include <stdio.h>
#include <ctype.h>
#include <gc.h>
#include <stdarg.h>
#include "Str.h"

#define INITIAL_STR_SIZE 32

Str
Strnew()
{
  Str             x = GC_malloc(sizeof(struct _Str));
  x->ptr = GC_malloc_atomic(INITIAL_STR_SIZE);
  x->area_size = INITIAL_STR_SIZE;
  x->length = 0;
  return x;
}

Str
Strnew_size(int n)
{
  Str             x = GC_malloc(sizeof(struct _Str));
  x->ptr = GC_malloc_atomic(n + 1);
  x->area_size = n + 1;
  x->length = 0;
  return x;
}

Str
Strnew_charp(char *p)
{
  Str             x = GC_malloc(sizeof(struct _Str));
  int             n;

  if (p == NULL)
    return Strnew();
  n = strlen(p) + 1;
  x->ptr = GC_malloc_atomic(n);
  x->area_size = n;
  x->length = n - 1;
  strcpy(x->ptr, p);
  return x;
}

Str
Strnew_charp_n(char *p, int n)
{
  Str             x = GC_malloc(sizeof(struct _Str));
  int             len;

  if (p == NULL)
    return Strnew_size(n);
  len = strlen(p);
  if (len < n)
    n = len;
  x->ptr = GC_malloc_atomic(n);
  x->area_size = n;
  x->length = n - 1;
  strncpy(x->ptr, p, n);
  return x;
}

Str
Strdup(Str s)
{
  Str             n = Strnew_size(s->length);
  Strcopy(n, s);
  return n;
}

void
Strclear(Str s)
{
  s->length = 0;
  s->ptr[0] = '\0';
}

void
Strfree(Str x)
{
  GC_free(x->ptr);
  GC_free(x);
}

void
Strcopy(Str x, Str y)
{
  if (x->area_size < y->length + 1) {
    GC_free(x->ptr);
    x->ptr = GC_malloc_atomic(y->length + 1);
    x->area_size = y->length + 1;
  }
  strcpy(x->ptr, y->ptr);
  x->length = y->length;
}

void
Strcopy_charp(Str x, char *y)
{
  int             len;
  if (y == NULL) {
    x->length = 0;
    return;
  }
  len = strlen(y);
  if (x->area_size < len + 1) {
    GC_free(x->ptr);
    x->ptr = GC_malloc_atomic(len + 1);
    x->area_size = len + 1;
  }
  strcpy(x->ptr, y);
  x->length = len;
}

void
Strcopy_charp_n(Str x, char *y, int n)
{
  int             len = n;

  if (y == NULL) {
    x->length = 0;
    return;
  }
  if (x->area_size < len + 1) {
    GC_free(x->ptr);
    x->ptr = GC_malloc_atomic(len + 1);
    x->area_size = len + 1;
  }
  strncpy(x->ptr, y, n);
  len = strlen(y);
  if (len < n)
    n = len;
  x->length = n;
}

void
Strcat(Str x, Str y)
{
  int newlen = x->length + y->length + 1;
  char *old = NULL;
  if (x->area_size < newlen) {
    old = x->ptr;
    newlen = newlen * 3 / 2;
    x->ptr = GC_malloc_atomic(newlen);
    x->area_size = newlen;
    strcpy(x->ptr,old);
  }
  strcpy(&x->ptr[x->length], y->ptr);
  x->length += y->length;
  if (old)
    GC_free(old);
}

void
Strcat_charp(Str x, char *y)
{
  int len,newlen;

  if (y == NULL)
    return;
  len = strlen(y);
  newlen = x->length + len + 1;
  if (x->area_size < newlen) {
    char           *old = x->ptr;
    newlen = newlen * 3 / 2;
    x->ptr = GC_malloc_atomic(newlen);
    x->area_size = newlen;
    strcpy(x->ptr, old);
    GC_free(old);
  }
  strcpy(&x->ptr[x->length], y);
  x->length += len;
}

void
Strcat_char(Str x, char y)
{
  int             newlen = x->length + 2;

  if (y == '\0')
    return;
  if (x->area_size < newlen) {
    char           *old = x->ptr;
    newlen = newlen * 3 / 2;
    x->ptr = GC_malloc_atomic(newlen);
    x->area_size = newlen;
    strcpy(x->ptr, old);
    GC_free(old);
  }
  x->ptr[x->length] = y;
  x->length++;
  x->ptr[x->length] = '\0';
}

Str
Strsubstr(Str s, int beg, int len)
{
  Str             new_s;
  int             i;
  if (beg >= s->length)
    return new_s;
  new_s = Strnew_size(len);
  if (len <= 0)
    len = s->length - beg;
  for (i = 0; i < len && beg + i < s->length; i++)
    Strcat_char(new_s, s->ptr[i]);
  return new_s;
}

void
Strlower(Str s)
{
  int             i;
  for (i = 0; i < s->length; i++)
    s->ptr[i] = tolower(s->ptr[i]);
}

void
Strupper(Str s)
{
  int             i;
  for (i = 0; i < s->length; i++)
    s->ptr[i] = tolower(s->ptr[i]);
}

void
Strchop(Str s)
{
  while ((s->ptr[s->length - 1] == '\n' || s->ptr[s->length - 1] == '\r') &&
	 s->length >= 0) {
    s->length--;
  }
  s->ptr[s->length] = '\0';
}

void
Strshrink(Str s, int n)
{
  if (n >= s->length) {
    s->length = 0;
    s->ptr[0] = '\0';
  } else {
    s->length -= n;
    s->ptr[s->length] = '\0';
  }
}

void
Strshrinkfirst(Str s, int n)
{
  int i;
  if (n >= s->length) {
    s->length = 0;
    s->ptr[0] = '\0';
  } else {
    for (i = n; i <= s->length; i++)
      s->ptr[i-n] = s->ptr[i];
    s->length -= n;
  }
}
  

void
Strremovefirstspaces(Str s)
{
  int             i, j;
  char           *old;

  for (i = 0; i < s->length && isspace(s->ptr[i]); i++);
  if (i == 0)
    return;
  s->length -= i;
  for (j = 0; j < s->length; j++)
    s->ptr[j] = s->ptr[j + i];
  s->ptr[j] = '\0';
}

void
Strremovetrailingspaces(Str s)
{
  int             i, j;
  char           *old;

  for (i = s->length - 1; i >= 0 && isspace(s->ptr[i]); i--);
  s->length = i + 1;
  s->ptr[i + 1] = '\0';
}

Str
Stralign_left(Str s, int width)
{
  Str             n;
  int             i;

  if (s->length >= width)
    return Strdup(s);
  n = Strnew_size(width);
  Strcopy(n, s);
  for (i = s->length; i < width; i++)
    Strcat_char(n, ' ');
  return n;
}

Str
Stralign_right(Str s, int width)
{
  Str             n;
  int             i;

  if (s->length >= width)
    return Strdup(s);
  n = Strnew_size(width);
  for (i = s->length; i < width; i++)
    Strcat_char(n, ' ');
  Strcat(n, s);
  return n;
}

Str
Stralign_center(Str s, int width)
{
  Str             n;
  int             i, w;

  if (s->length >= width)
    return Strdup(s);
  n = Strnew_size(width);
  w = (width - s->length) / 2;
  for (i = 0; i < w; i++)
    Strcat_char(n, ' ');
  Strcat(n, s);
  for (i = w + s->length; i < width; i++)
    Strcat_char(n, ' ');
  return n;
}

#define SP_NORMAL 0
#define SP_PREC   1
#define SP_PREC2  2

Str
Sprintf(char *fmt,...)
{
  int             len = 0;
  int             status = SP_NORMAL;
  int             p;
  char           *f;
  Str             s;
  va_list         ap;

  for (f = fmt; *f; f++) {
    switch (status) {
    case SP_NORMAL:
      if (*f == '%') {
	status = SP_PREC;
	p = 0;
      } else
	len++;
      break;
    case SP_PREC:
      if (isalpha(*f)) {
	/* conversion char. */
	if (p > 0) {
	  /* precision specified */
	  len += p;
	} else {
	  len += 20;		/* arbitrary */
	}
	status = SP_NORMAL;
      } else if (isdigit(*f))
	p = p * 10 + *f - '0';
      else if (*f == '.')
	status = SP_PREC2;
      break;
    case SP_PREC2:
      if (isalpha(*f)) {
	/* conversion char. */
	if (p > 0) {
	  /* precision specified */
	  len += p;
	} else {
	  len += 20;		/* arbitrary */
	}
	status = SP_NORMAL;
      }
      break;
    }
  }
  s = Strnew_size(len * 2);
  va_start(ap, fmt);
  vsprintf(s->ptr, fmt, ap);
  va_end(ap);
  s->length = strlen(s->ptr);
  return s;
}

Str
Strfgets(FILE *f)
{
  Str s = Strnew();
  char c;
  while (1) {
    c = fgetc(f);
    if (feof(f)) break;
    Strcat_char(s,c);
    if (c == '\n') break;
  }
  return s;
}

Str
Strfgetall(FILE *f)
{
  Str s = Strnew();
  char c;
  while (1) {
    c = fgetc(f);
    if (feof(f)) break;
    Strcat_char(s,c);
  }
  return s;
}
